pub mod audiowrapper {

    use std::thread::*;
    use std::error::Error;
    use std::time::Duration;
    use std::io::BufReader;
    use std::fs::File;
    use std::any::Any;

    use flume::{Receiver, Sender};
    use rodio::{Decoder, OutputStream, Sink, Source};
    use std::path::{Path, PathBuf};

    use crate::playlist::playlist::*;
    use crate::songmeta::songmeta::*;


    //A wrapper around a rodio sink that allows for better controls.
    //Can't use sink queues so I implement myself...
    //Reads songs in threads, one at a time so we can stop, play, pause, insert songs in playlist, etc.
    #[derive(Debug)]
    pub struct AudioWrapper {
        tx: Sender<ThreadCom>,
        rx: Receiver<ThreadCom>,
        pub playlist: Playlist,
        handle: Option<JoinHandle<()>>,
    }

    //Used for handling communication with the thread
    enum ThreadCom {
        Stop,
        Pause,
        Unpause,
        SendPos,
        Pos(Duration),
        SetVol(f32),
        GetVol,
    }

    impl AudioWrapper {
        pub fn new() -> Self {
            let (tx, rx) = flume::unbounded();
            Self {
                tx: tx,
                rx: rx,
                playlist: Playlist::new(),
                handle: None,
            }
        }

        //Inner function used in a thread to read music
        fn sub_sink(path: String, tx: Sender<ThreadCom>, rx: Receiver<ThreadCom>) {
            let path = Path::new(&path);
            let (stream, stream_handle) = OutputStream::try_default().unwrap();
            let sink = Sink::try_new(&stream_handle).unwrap();
            let file = BufReader::new(File::open(&path).unwrap());
            let source = Decoder::new(file).unwrap();
            sink.append(source);

            loop {

                let recv = rx.try_recv();

                match recv {
                    Ok(v) => {
                        match v {
                            ThreadCom::Stop => {
                                sink.stop();
                                break;
                            },
                            ThreadCom::Pause => {
                                sink.pause();
                            },
                            ThreadCom::Unpause => {
                                sink.play();
                            },
                            ThreadCom::SendPos => {
                                tx.try_send(ThreadCom::Pos(sink.get_pos()));
                            },
                            ThreadCom::SetVol(w) => {
                                sink.set_volume(w);
                            },
                            ThreadCom::GetVol => {
                                tx.try_send(ThreadCom::SetVol(sink.volume()));
                            },
                            ThreadCom::Pos(w) => {
                                let _ = sink.try_seek(w);
                            }
                        };
                    },
                    Err(_) => (),
                };

                //Finished
                if sink.len() < 1 {
                    break;
                }
            }
        }

        ///Reads ONE song by starting a sink in a new thread.
        ///Stop any previous song playing.
        pub fn play_at(&mut self, index: usize){
            self.playlist.check_index(index.clone()).unwrap();
            self.stop();
            let path = self.playlist.songs[index].path.clone();
            let tx2 = self.tx.clone();
            let rx2 = self.rx.clone();
            let handle = spawn(|| {
                AudioWrapper::sub_sink(path, tx2, rx2);
            });
            self.handle = Some(handle);
        }

        pub fn stop(&mut self) {
            match &self.handle {
                Some(h) => {
                    self.tx.try_send(ThreadCom::Stop).unwrap();
                    self.handle.take().expect("Called stop on non-running thread").join().expect("Could not join");
                },
                None => (),
            };
        }
    }
}